
/*
 * COPYRIGHT NOTICE: 
 * Copyright (c) Tim Bray, 1990.
 * Copyright (c) Russell Coker, 1999.  I have updated the program, added
 * support for >2G on 32bit machines, and tests for file creation.  Same
 * license as Tim's original code.
 * Everybody is hereby granted rights to use, copy, and modify this program, 
 *  provided only that this copyright notice and the disclaimer below
 *  are preserved without change.
 * DISCLAIMER:
 * This program is provided AS IS with no warranty of any kind, and
 * The author makes no representation with respect to the adequacy of this
 *  program for any particular purpose or with respect to its adequacy to 
 *  produce any particular result, and
 * The author shall not be liable for loss or damage arising out of
 *  the use of this program regardless of how sustained, and
 * In no event shall the author be liable for special, direct, indirect
 *  or consequential damage, loss, costs or fees or expenses of any
 *  nature or kind.
 */

#include "bonnie.h"
#include <iostream.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <stdlib.h>
#include <unistd.h>
#include "fileio.h"
#include "bon_time.h"
#include <ctype.h>
#include <string.h>

void usage();

class GlobalItems
{
public:
  bool quiet;
  BonTimer timer;
  int ram;
  char *name;
  GlobalItems()
  {
    quiet = false;
    ram = 0;
    name = NULL;
    SetName(".");
  }
  ~GlobalItems() { delete name; }

  void SetName(CPCCHAR path)
  {
    delete name;
    name = new char[strlen(path) + 15];
    sprintf(name, "%s/Bonnie.%d", path, int(getpid()));
  }
};

int TestDirOps(int directory_size, GlobalItems &globals);

int main(int argc, char *argv[])
{
  int    buf[Chunk / IntSize];
  int    bufindex;
  int    chars[256];
  int    child;
  int    lseek_count = 0;
  int    next;
  int    seek_control[2];
  int    seek_feedback[2];
  char   seek_tickets[Seeks + SeekProcCount];
  int    file_size = DefaultFileSize;
  int    directory_size = DefaultDirectorySize;
  int    num_chunks;
  int    words;
  CFileOp file;
  int i;
  char * machine = "Unknown";
  GlobalItems globals;

  for (next = 1; next < argc; next++)
  {
    if (argv[next][0] == '-')
    { /* option? */
      switch(tolower(argv[next][1]))
      {
      case 'd':
        if(next + 1 == argc)
          usage();
        globals.SetName(argv[next + 1]);
      break;
      case 's':
        if(next + 1 == argc)
          usage();
        file_size = atoi(argv[next + 1]);
      break;
      case 'n':
        if(next + 1 == argc)
          usage();
        directory_size = atoi(argv[next + 1]);
      break;
      case 'm':
        machine = argv[next + 1];
      break;
      case 'r':
        if(next + 1 == argc)
          usage();
        globals.ram = atoi(argv[next + 1]);
      break;
      case 't':
        if(strcmp(argv[next + 1], "csv") == 0)
        {
          globals.timer.SetType(BonTimer::csv);
          globals.quiet = true;
        }
      break;
      default:
        usage();
      }
      next++;
    } /* option? */
    else
      usage();
  }

  if (file_size < 0 || directory_size < 0 || (!file_size && !directory_size) )
    usage();

  if(file_size)
  {
    if(globals.ram && file_size < globals.ram * 2)
    {
      cerr << "File size should be double RAM for good results.\n";
      return 1;
    }
    file.set_size(file_size);
    // default is we have 1M / 8K * 100 chunks = 12800
    num_chunks = Unit / Chunk * file_size;

    file.open(globals.name, true);
    globals.timer.timestamp();

    /* Fill up a file, writing it a char at a time with the stdio putc() call */
    if(!globals.quiet) cerr << "Writing with putc()...";
    for(words = 0; words < num_chunks; words++)
    {
      if(file.write_block_putc() == -1)
        return 1;
    }

    /*
     * note that we always close the file before measuring time, in an
     *  effort to force as much of the I/O out as we can
     */
    file.close();
    globals.timer.get_delta_t(Putc);
    if(!globals.quiet) cerr << "done\n";

    /* Now read & rewrite it using block I/O.  Dirty one word in each block */
    if(file.reopen(false))
      return 1;
    if (file.seek(0, SEEK_SET) == (off_t) -1)
    {
      if(!globals.quiet) cerr << "error in lseek(2) before rewrite\n";
      return 1;
    }
    if(!globals.quiet) cerr << "Rewriting...";
    globals.timer.timestamp();
    bufindex = 0;
    if ((words = file.read_block(PVOID(buf))) == -1)
      return io_error("rewrite read");
    while (words == Chunk)
    { /* while we can read a block */
      if (bufindex == Chunk / IntSize)
        bufindex = 0;
      buf[bufindex++]++;
      if (file.seek(-1, SEEK_CUR) == -1)
        return io_error("relative lseek(2)");
      if (file.write_block(PVOID(buf)) == -1)
        return io_error("re write(2)");
      if ((words = file.read_block(PVOID(buf))) == -1)
        return io_error("rwrite read");
    } /* while we can read a block */
    file.close();
    globals.timer.get_delta_t(ReWrite);
    if(!globals.quiet) cerr << "done\n";

    /* Write the whole file from scratch, again, with block I/O */
    if(file.reopen(true))
      return true;
    if(!globals.quiet) cerr << "Writing intelligently...";
    for (words = 0; words < int(Chunk / IntSize); words++)
      buf[words] = 0;
    globals.timer.timestamp();
    for(i = 0; i < file_size; i++)
    {
      for (words = bufindex = 0; words < (Unit / Chunk); words++)
      { /* for each word */
        if (bufindex == (Chunk / IntSize))
          bufindex = 0;
        buf[bufindex++]++;
        if (file.write_block(PVOID(buf)) == -1)
          return io_error("write(2)");
      } /* for each word */
    }
    file.close();
    globals.timer.get_delta_t(FastWrite);
    if(!globals.quiet) cerr << "done\n";

    /* read them all back with getc() */
    if(file.reopen(false))
      return 1;
    for (words = 0; words < 256; words++)
      chars[words] = 0;
    if(!globals.quiet) cerr << "Reading with getc()...";
    globals.timer.timestamp();

    for(words = 0; words < num_chunks; words++)
    {
      if(file.read_block_getc((char *)chars) == -1)
        return 1;
    }

    file.close();
    globals.timer.get_delta_t(Getc);
    if(!globals.quiet) cerr << "done\n";

    /* use the frequency count */
    for (words = 0; words < 256; words++)
      sprintf((char *) buf, "%d", chars[words]);

    /* Now suck it in, Chunk at a time, as fast as we can */
    if(file.reopen(false))
      return 1;
    if (file.seek(0, SEEK_SET) == -1)
      return io_error("lseek before read");
    if(!globals.quiet) cerr << "Reading intelligently...";
    globals.timer.timestamp();
    do
    { /* per block */
      if ((words = file.read_block(PVOID(buf))) == -1)
        return io_error("read(2)");
      chars[buf[abs(buf[0]) % (Chunk / IntSize)] & 0x7f]++;
    } /* per block */
    while (words);
    file.close();
    globals.timer.get_delta_t(FastRead);
    if(!globals.quiet) cerr << "done\n";

    /* use the frequency count */
    for (words = 0; words < 256; words++)
      sprintf((char *) buf, "%d", chars[words]);

    /*
     * Now test random seeks; first, set up for communicating with children.
     * The object of the game is to do "Seeks" lseek() calls as quickly
     *  as possible.  So we'll farm them out among SeekProcCount processes.
     *  We'll control them by writing 1-byte tickets down a pipe which
     *  the children all read.  We write "Seeks" bytes with val 1, whichever
     *  child happens to get them does it and the right number of seeks get
     *  done.
     * The idea is that since the write() of the tickets is probably
     *  atomic, the parent process likely won't get scheduled while the
     *  children are seeking away.  If you draw a picture of the likely
     *  timelines for three children, it seems likely that the seeks will
     *  overlap very nicely with the process scheduling with the effect
     *  that there will *always* be a seek() outstanding on the file.
     * Question: should the file be opened *before* the fork, so that
     *  all the children are lseeking on the same underlying file object?
     */
    if (pipe(seek_feedback) == -1 || pipe(seek_control) == -1)
      return io_error("pipe");
    for (next = 0; next < Seeks; next++)
      seek_tickets[next] = 1;
    for ( ; next < (Seeks + SeekProcCount); next++)
      seek_tickets[next] = 0;

    /* launch some parallel seek processes */
    for (next = 0; next < SeekProcCount; next++)
    { /* for each seek proc */
      struct report_s seeker_report;

      if ((child = fork()) == -1)
        return io_error("fork");
      else if (child == 0)
      { /* child process */

        /* set up and wait for the go-ahead */
        close(seek_feedback[0]);
        close(seek_control[1]);
        if(file.reopen(false))
          return 1;
        srandom(getpid());
        if(!globals.quiet) cerr << "Seeker " << next << "...";

        /* wait for the go-ahead */
        if (read(seek_control[0], seek_tickets, 1) != 1)
	  return io_error("read ticket");
        globals.timer.timestamp();
        seeker_report.StartTime = globals.timer.time_so_far();

        /* loop until we read a 0 ticket back from our parent */
        while(seek_tickets[0])
        { /* until Mom says stop */
          if(file.doseek(random() % num_chunks,
	    ((lseek_count++ % UpdateSeek) == 0)))
            return 1;
	  if (read(seek_control[0], seek_tickets, 1) != 1)
	    return io_error("read ticket");
        } /* until Mom says stop */
        file.close();

        /* report to parent */
        globals.timer.get_delta_report(seeker_report);
        if (write(seek_feedback[1], (const void *)&seeker_report, sizeof(seeker_report))
            != sizeof(seeker_report))
          return io_error("pipe write");
        exit(0);
      } /* child process */
    } /* for each seek proc */

    /*
     * Back in the parent; in an effort to ensure the children get an even
     *  start, wait a few seconds for them to get scheduled, open their
     *  files & so on.
     */
    close(seek_feedback[1]);
    close(seek_control[0]);
    sleep(5);
    if(!globals.quiet) cerr << "start 'em...";
    if (write(seek_control[1], seek_tickets, sizeof(seek_tickets)) 
        != sizeof(seek_tickets))
      return io_error("write tickets");

    /* read back from children */
    for (next = 0; next < SeekProcCount; next++)
    { /* for each child */
      struct report_s seeker_report;

      if (read(seek_feedback[0], (void *)&seeker_report, sizeof(seeker_report))
          != sizeof(seeker_report))
        return io_error("pipe read");

      /*
       * each child writes back its CPU, start & end times.  The elapsed time 
       *  to do all the seeks is the time the first child started until the 
       *  time the last child stopped
       */
      globals.timer.add_delta_report(seeker_report, Lseek);
      if (wait(&child) == -1)
        return io_error("wait");
      if(!globals.quiet) cerr << "done...";
    } /* for each child */
    if(!globals.quiet) cerr << endl;
  }

  TestDirOps(directory_size, globals);

  globals.timer.DoReport(machine, file_size, directory_size);
}

int
TestDirOps(int directory_size, GlobalItems &globals)
{
  COpenTest open_test;
  if(directory_size)
  {
    if(!globals.quiet) cerr << "Create files in sequential order...";
    if(open_test.create(globals.name, globals.timer, directory_size, false))
      return 1;
    if(!globals.quiet) cerr << "done.\nStat files in sequential order...";
    if(open_test.stat_sequential(globals.timer))
      return 1;
    if(!globals.quiet) cerr << "done.\nDelete files in sequential order...";
    if(open_test.delete_sequential(globals.timer))
      return 1;
    if(!globals.quiet) cerr << "done.\n";

    if(globals.ram && directory_size * MaxNameLen > (globals.ram << 20))
    {
      cerr << "When testing " << directory_size << " files in " << globals.ram
           << "MB of RAM the system is likely to\n"
              "start paging Bonnie++ data and the test will give suspect\n"
              "results, use less files or have more RAM for this test.\n";
    }
    else
    {
      if(!globals.quiet) cerr << "Create files in random order...";
      if(open_test.create(globals.name, globals.timer, directory_size, true))
        return 1;
      if(!globals.quiet) cerr << "done.\nStat files in random order...";
      if(open_test.stat_random(globals.timer))
        return 1;
      if(!globals.quiet) cerr << "done.\nDelete files in random order...";
      if(open_test.delete_random(globals.timer))
        return 1;
      if(!globals.quiet) cerr << "done.\n";
    }
  }
  return 0;
}

void
usage()
{
  fprintf(stderr,
    "usage: Bonnie [-d scratch-dir] [-s size-in-Mb]\n"
    "              [-n number-to-stat] [-m machine-label]\n"
    "              [-t csv] for CSV output\n"
    "              [-r ram-size-in-Mb]\n"
    "\nVersion: " BON_VERSION "\n");
  exit(1);
}

int
io_error(CPCCHAR message, bool do_exit)
{
  char buf[Chunk];

  sprintf(buf, "Bonnie: drastic I/O error (%s)", message);
  perror(buf);
  if(do_exit)
    exit(1);
  return(1);
}

#ifdef SysV
static char randseed[32];
 
void
srandom(int seed)
{
  sprintf(randseed, "%06d", seed);
}

long
random()
{
  return nrand48(randseed);
}
#endif

